/*
 * ctrl.c
 *
 * A set of routines for processing quake control packets
 *
 * Kekoa Proudfoot 12/30/96
 */

#include <stdlib.h>
#include <stdio.h>
#include "ctrl.h"
#include "endian.h"
#include "hexdump.h"

/*****************************************************************************
 * Macro definitions
 *****************************************************************************/

/* Get macros */

#define GET_CHAR(b)          (b += 1, b[-1])
#define GET_SHORT(b)         (b += 2, read_little_short(&b[-2]))
#define GET_LONG(b)          (b += 4, read_little_long (&b[-4]))

#define GET_STRING(b)        b; while (*b++)

/* Put macros */

#define PUT_CHAR(b,v)        (b[0] = v,                 b += 1)
#define PUT_SHORT(b,v)       (write_little_short(b, v), b += 2)
#define PUT_LONG(b,v)        (write_little_long (b, v), b += 4)

#define PUT_STRING(b,v)      { char *c = v; do { *b = *c++; } while (*b++); }

/*****************************************************************************
 * Interface function definitions
 *****************************************************************************/

int
cl_to_sv_ctrl_get(char *buf, cl_to_sv_ctrl_t *ctrl)
{
    char  *start = buf;

    switch (*buf) {
    case CTRL_REQ_CONNECT:
	ctrl->req_connect.op              = GET_CHAR(buf);
	ctrl->req_connect.game            = GET_STRING(buf);
	ctrl->req_connect.net_version     = GET_CHAR(buf);
	break;
    case CTRL_REQ_SERVER_INFO:
	ctrl->req_server_info.op          = GET_CHAR(buf);
	ctrl->req_server_info.game        = GET_STRING(buf);
	ctrl->req_server_info.net_version = GET_CHAR(buf);
	break;
    case CTRL_REQ_PLAYER_INFO:
	ctrl->req_player_info.op          = GET_CHAR(buf);
	ctrl->req_player_info.player      = GET_CHAR(buf);
	break;
    case CTRL_REQ_RULE_INFO:
	ctrl->req_rule_info.op            = GET_CHAR(buf);
	ctrl->req_rule_info.rule          = GET_STRING(buf);
	break;
    default:
	ctrl->op                          = CTRL_BAD;
	break;
    }

    return buf - start;
}

int
cl_to_sv_ctrl_put(char *buf, cl_to_sv_ctrl_t *ctrl)
{
    char  *start = buf;

    switch (ctrl->op) {
    case CTRL_REQ_CONNECT:
	PUT_CHAR(buf, ctrl->req_connect.op);
	PUT_STRING(buf, ctrl->req_connect.game);
	PUT_CHAR(buf, ctrl->req_connect.net_version);
	break;
    case CTRL_REQ_SERVER_INFO:
	PUT_CHAR(buf, ctrl->req_server_info.op);
	PUT_STRING(buf, ctrl->req_server_info.game);
	PUT_CHAR(buf, ctrl->req_server_info.net_version);
	break;
    case CTRL_REQ_PLAYER_INFO:
	PUT_CHAR(buf, ctrl->req_player_info.op);
	PUT_CHAR(buf, ctrl->req_player_info.player);
	break;
    case CTRL_REQ_RULE_INFO:
	PUT_CHAR(buf, ctrl->req_rule_info.op);
	PUT_STRING(buf, ctrl->req_rule_info.rule);
	break;
    default:
	PUT_CHAR(buf, CTRL_BAD);
	break;
    }

    return buf - start;
}

char *
cl_to_sv_ctrl_name(cl_to_sv_ctrl_t *ctrl)
{
    switch (ctrl->op) {
    default:
    case CTRL_BAD:
	return "CTRL_BAD";
    case CTRL_REQ_CONNECT:
	return "CTRL_REQ_CONNECT";
    case CTRL_REQ_SERVER_INFO:
	return "CTRL_REQ_SERVER_INFO";
    case CTRL_REQ_PLAYER_INFO:
	return "CTRL_REQ_PLAYER_INFO";
    case CTRL_REQ_RULE_INFO:
	return "CTRL_REQ_RULE_INFO";
    }
}

void
cl_to_sv_ctrl_debug(cl_to_sv_ctrl_t *ctrl, FILE *file)
{
    char buf[16384];
    int  len;

    switch (ctrl->op) {
    default:
	len = cl_to_sv_ctrl_put(buf, ctrl);
	fprintf(file, "[client control: %s, %d bytes]\n",
		cl_to_sv_ctrl_name(ctrl), len);
	hex_dump(buf, len, "C", file);
	break;
    }
}

char
cl_to_sv_ctrl_peek(char *buf)
{
    switch (*buf) {
    default:
    case CTRL_BAD:
	return CTRL_BAD;
    case CTRL_REQ_CONNECT:
    case CTRL_REQ_SERVER_INFO:
    case CTRL_REQ_PLAYER_INFO:
    case CTRL_REQ_RULE_INFO:
	return *buf;
    }
}

int
sv_to_cl_ctrl_get(char *buf, sv_to_cl_ctrl_t *ctrl)
{
    char  *start = buf;
    char player_color;

    switch (*buf) {
    case CTRL_REP_ACCEPT:
	ctrl->rep_accept.op               = GET_CHAR(buf);
	ctrl->rep_accept.port             = GET_SHORT(buf);
	ctrl->rep_accept.unknown          = GET_SHORT(buf);
	break;
    case CTRL_REP_REJECT:
	ctrl->rep_reject.op               = GET_CHAR(buf);
	ctrl->rep_reject.reason           = GET_STRING(buf);
	break;
    case CTRL_REP_SERVER_INFO:
	ctrl->rep_server_info.op          = GET_CHAR(buf);
	ctrl->rep_server_info.address     = GET_STRING(buf);
	ctrl->rep_server_info.host        = GET_STRING(buf);
	ctrl->rep_server_info.map         = GET_STRING(buf);
	ctrl->rep_server_info.num_clients = GET_CHAR(buf);
	ctrl->rep_server_info.max_clients = GET_CHAR(buf);
	ctrl->rep_server_info.net_version = GET_CHAR(buf);
	break;
    case CTRL_REP_PLAYER_INFO:
	ctrl->rep_player_info.op          = GET_CHAR(buf);
	ctrl->rep_player_info.player      = GET_CHAR(buf);
	ctrl->rep_player_info.name        = GET_STRING(buf);
	player_color                      = GET_CHAR(buf);
	ctrl->rep_player_info.shirt_color = (player_color >> 4) & 0x0f;
	ctrl->rep_player_info.pants_color = (player_color >> 0) & 0x0f;
	ctrl->rep_player_info.unknown[0]  = GET_CHAR(buf);
	ctrl->rep_player_info.unknown[1]  = GET_CHAR(buf);
	ctrl->rep_player_info.unknown[2]  = GET_CHAR(buf);
	ctrl->rep_player_info.frags       = GET_LONG(buf);
	ctrl->rep_player_info.time        = GET_LONG(buf);
	ctrl->rep_player_info.address     = GET_STRING(buf);
	break;
    case CTRL_REP_RULE_INFO:
	ctrl->rep_rule_info.op            = GET_CHAR(buf);
	ctrl->rep_rule_info.rule          = GET_STRING(buf);
	ctrl->rep_rule_info.value         = GET_STRING(buf);
	break;
    default:
	ctrl->op                          = CTRL_BAD;
	break;
    }

    return buf - start;
}

int
sv_to_cl_ctrl_put(char *buf, sv_to_cl_ctrl_t *ctrl)
{
    char  *start = buf;

    switch (ctrl->op) {
    case CTRL_REP_ACCEPT:
	PUT_CHAR(buf, ctrl->rep_accept.op);
	PUT_SHORT(buf, ctrl->rep_accept.port);
	PUT_SHORT(buf, ctrl->rep_accept.unknown);
	break;
    case CTRL_REP_REJECT:
	PUT_CHAR(buf, ctrl->rep_reject.op);
	PUT_STRING(buf, ctrl->rep_reject.reason);
	break;
    case CTRL_REP_SERVER_INFO:
	PUT_CHAR(buf, ctrl->rep_server_info.op);
	PUT_STRING(buf, ctrl->rep_server_info.address);
	PUT_STRING(buf, ctrl->rep_server_info.host);
	PUT_STRING(buf, ctrl->rep_server_info.map);
	PUT_CHAR(buf, ctrl->rep_server_info.num_clients);
	PUT_CHAR(buf, ctrl->rep_server_info.max_clients);
	PUT_CHAR(buf, ctrl->rep_server_info.net_version);
	break;
    case CTRL_REP_PLAYER_INFO:
	PUT_CHAR(buf, ctrl->rep_player_info.op);
	PUT_CHAR(buf, ctrl->rep_player_info.player);
	PUT_STRING(buf, ctrl->rep_player_info.name);
	PUT_CHAR(buf, ((ctrl->rep_player_info.shirt_color & 0x0f) << 4) |
		      ((ctrl->rep_player_info.pants_color & 0x0f) << 0));
	PUT_CHAR(buf, ctrl->rep_player_info.unknown[0]);
	PUT_CHAR(buf, ctrl->rep_player_info.unknown[1]);
	PUT_CHAR(buf, ctrl->rep_player_info.unknown[2]);
	PUT_LONG(buf, ctrl->rep_player_info.frags);
	PUT_LONG(buf, ctrl->rep_player_info.time);
	PUT_STRING(buf, ctrl->rep_player_info.address);
	break;
    case CTRL_REP_RULE_INFO:
	PUT_CHAR(buf, ctrl->rep_rule_info.op);
	PUT_STRING(buf, ctrl->rep_rule_info.rule);
	PUT_STRING(buf, ctrl->rep_rule_info.value);
	break;
    default:
	PUT_CHAR(buf, CTRL_BAD);
	break;
    }

    return buf - start;
}

char *
sv_to_cl_ctrl_name(sv_to_cl_ctrl_t *ctrl)
{
    switch (ctrl->op) {
    default:
    case CTRL_BAD:
	return "CTRL_BAD";
    case CTRL_REP_ACCEPT:
	return "CTRL_REP_ACCEPT";
    case CTRL_REP_REJECT:
	return "CTRL_REP_REJECT";
    case CTRL_REP_SERVER_INFO:
	return "CTRL_REP_SERVER_INFO";
    case CTRL_REP_PLAYER_INFO:
	return "CTRL_REP_PLAYER_INFO";
    case CTRL_REP_RULE_INFO:
	return "CTRL_REP_RULE_INFO";
    }
}

void
sv_to_cl_ctrl_debug(sv_to_cl_ctrl_t *ctrl, FILE *file)
{
    char buf[16384];
    int  len;

    switch (ctrl->op) {
    default:
	len = sv_to_cl_ctrl_put(buf, ctrl);
	fprintf(file, "[server control: %s, %d bytes]\n",
		sv_to_cl_ctrl_name(ctrl), len);
	hex_dump(buf, len, "S", file);
	break;
    }
}

char
sv_to_cl_ctrl_peek(char *buf)
{
    switch (*buf) {
    default:
    case CTRL_BAD:
	return CTRL_BAD;
    case CTRL_REP_ACCEPT:
    case CTRL_REP_REJECT:
    case CTRL_REP_SERVER_INFO:
    case CTRL_REP_PLAYER_INFO:
    case CTRL_REP_RULE_INFO:
	return *buf;
    }
}
