/*
 * packet.c
 *
 * Packet routines.
 *
 * Kekoa Proudfoot 1/1/97
 */

/*****************************************************************************
 * Include files
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "packet.h"
#include "mesg.h"
#include "ctrl.h"
#include "debug.h"
#include "hexdump.h"
#include "endian.h"

/*****************************************************************************
 * Interface function definitions
 *****************************************************************************/

packet_t *
packet_create(unsigned short type)
{
    int        size;
    packet_t  *packet;

    switch (type) {
    default:
    case PACKET_UNDEFINED:
    case PACKET_CONTROL:
    case PACKET_UNRELIABLE:
    case PACKET_RELIABLE:
    case PACKET_RELIABLE_REPLY:
    case PACKET_RELIABLE_MORE:
    case PACKET_RELIABLE_END:
	size = sizeof(packet_t);
	break;
    }

    if ((packet = (packet_t *)malloc(sizeof(packet_t))) == NULL) {
	debug_fatal("packet_create: malloc error");
    }

    packet->header.type   = type;
    packet->header.length = 0;

    return packet;
}

void
packet_destroy(packet_t *packet)
{
    if (packet)
	free(packet);
}

void
packet_copy(packet_t *dst, packet_t *src)
{
    switch (src->type) {
    case PACKET_UNDEFINED:
    default:
	memcpy(dst, src, src->header.length);
	dst->header.type   = src->header.type;
	dst->header.length = src->header.length;
	break;
    case PACKET_CONTROL:
	memcpy(dst, src, src->header.length + 4);
	break;
    case PACKET_RELIABLE:
    case PACKET_UNRELIABLE:
    case PACKET_RELIABLE_REPLY:
    case PACKET_RELIABLE_MORE:
    case PACKET_RELIABLE_END:
	memcpy(dst, src, src->header.length + 8);
	break;
    }
}

char *
packet_name(packet_t *packet)
{
    switch (packet->type) {
    default:
    case PACKET_UNDEFINED:
	return "PACKET_UNDEFINED";
    case PACKET_CONTROL:
	return "PACKET_CONTROL";
    case PACKET_RELIABLE:
	return "PACKET_RELIABLE";
    case PACKET_UNRELIABLE:
	return "PACKET_UNRELIABLE";
    case PACKET_RELIABLE_REPLY:
	return "PACKET_RELIABLE_REPLY";
    case PACKET_RELIABLE_MORE:
	return "PACKET_RELIABLE_MORE";
    case PACKET_RELIABLE_END:
	return "PACKET_RELIABLE_END";
    }
}

void
packet_cl_to_sv_debug(packet_t *packet, FILE *file)
{
    int              position, length;
    cl_to_sv_ctrl_t  ctrl;
    cl_to_sv_mesg_t  mesg;

    switch (packet->type) {
    default:
    case PACKET_UNDEFINED:
	fprintf(file, "[client packet: %s, %d bytes]\n",
		packet_name(packet), packet->header.length);
	hex_dump(packet, packet->header.length, "C", file);
	break;
    case PACKET_CONTROL:
	fprintf(file, "[client packet: %s, %d bytes]\n",
		packet_name(packet), packet->control.length + 4);
	cl_to_sv_ctrl_get(packet->control.data, &ctrl);
	cl_to_sv_ctrl_debug(&ctrl, file);
	break;
    case PACKET_RELIABLE:
    case PACKET_UNRELIABLE:
    case PACKET_RELIABLE_REPLY:
    case PACKET_RELIABLE_MORE:
    case PACKET_RELIABLE_END:
	fprintf(file, "[client packet: %s, %d bytes, number %ld]\n",
		packet_name(packet), packet->header.length + 8,
		packet->header.number);
	position = 0; length = 1;
	while (length && position < packet->reliable.length) {
	    length = cl_to_sv_mesg_get(&packet->reliable.data[position],
				       &mesg);
	    cl_to_sv_mesg_debug(&mesg, file);
	    position += length;
	}
	break;
    }
}

void
packet_sv_to_cl_debug(packet_t *packet, FILE *file)
{
    int              position, length;
    sv_to_cl_ctrl_t  ctrl;
    sv_to_cl_mesg_t  mesg;

    switch (packet->type) {
    default:
    case PACKET_UNDEFINED:
	fprintf(file, "[server packet: %s, %d bytes]\n",
		packet_name(packet), packet->header.length);
	hex_dump(packet, packet->header.length, "S", file);
	break;
    case PACKET_CONTROL:
	fprintf(file, "[server packet: %s, %d bytes]\n",
		packet_name(packet), packet->control.length + 4);
	sv_to_cl_ctrl_get(packet->control.data, &ctrl);
	sv_to_cl_ctrl_debug(&ctrl, file);
	break;
    /* PACKET_RELIABLE_MORE are incomplete. For debugging purposes
       just list the length and number, the entire packet contents
       will be printed when the PACKET_RELIABLE_END comes along */
    case PACKET_RELIABLE_MORE:
	fprintf(file, "[server packet: %s, %d bytes, number %ld]\n",
		packet_name(packet), packet->header.length + 8,
		packet->header.number);
	break;
    case PACKET_RELIABLE:
    case PACKET_UNRELIABLE:
    case PACKET_RELIABLE_REPLY:
    case PACKET_RELIABLE_END:
	fprintf(file, "[server packet: %s, %d bytes, number %ld]\n",
		packet_name(packet), packet->header.length + 8,
		packet->header.number);
	position = 0; length = 1;
	while (length && position < packet->reliable.length) {
	    length = sv_to_cl_mesg_get(&packet->reliable.data[position],
				       &mesg);
	    sv_to_cl_mesg_debug(&mesg, file);
	    position += length;
	}
	break;
    }
}

void
packet_host_to_network(packet_t *p)
{
    switch (p->type) {
    case PACKET_UNDEFINED:
    default:
	write_big_short(&p->header.type,   p->header.type);
	write_big_short(&p->header.length, p->header.length);
	break;
    case PACKET_CONTROL:
	write_big_short(&p->header.type,   p->header.type);
	write_big_short(&p->header.length, p->header.length + 4);
	break;
    case PACKET_RELIABLE:
    case PACKET_UNRELIABLE:
    case PACKET_RELIABLE_REPLY:
    case PACKET_RELIABLE_MORE:
    case PACKET_RELIABLE_END:
	write_big_short(&p->header.type,   p->header.type);
	write_big_short(&p->header.length, p->header.length + 8);
	write_big_long (&p->header.number, p->header.number);
	break;
    }
}

void
packet_network_to_host(packet_t *p)
{
    p->header.type   = read_big_short(&p->header.type);
    switch (p->type) {
    case PACKET_UNDEFINED:
    default:
	p->header.length = read_big_short(&p->header.length);
	break;
    case PACKET_CONTROL:
	p->header.length = read_big_short(&p->header.length) - 4;
	break;
    case PACKET_RELIABLE:
    case PACKET_UNRELIABLE:
    case PACKET_RELIABLE_REPLY:
    case PACKET_RELIABLE_MORE:
    case PACKET_RELIABLE_END:
	p->header.length = read_big_short(&p->header.length) - 8;
	p->header.number = read_big_long (&p->header.number);
	break;
    }
}
